/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 * Copyright (C) 2007 Novell, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef E_FLAG_H
#define E_FLAG_H

/* An EFlag is essentially a binary semaphore with a more intuitive interface.
 * Based on Python's threading.Event class ("EEvent" was already taken). */

#include <glib.h>

G_BEGIN_DECLS

/* Opaque structure */
typedef struct _EFlag EFlag;

/*
 * Creates a new EFlag object.  It is initially unset.
 */
extern EFlag *e_flag_new(void);

/*
 * Returns TRUE if flag is set
 */
extern gboolean	e_flag_is_set(EFlag *flag);


/*
 * Sets flag. All threads waiting on flag are woken up. Threads that
 * call e_flag_wait() or e_flag_timed_wait() once flag is set will not
 * block at all.
 */
extern void e_flag_set(EFlag *flag);

/*
 * Unsets flag.  Subsequent calls to e_flag_wait() or e_flag_timed_wait()
 * will block until flag is set.
 */
extern void e_flag_clear(EFlag *flag);

/*
 * Blocks until flag is set. If flag is already set, the function returns
 * immediately.
 */
extern void e_flag_wait(EFlag *flag);

/*
 * Blocks until flag is set, or until the time specified by abs_time.
 * If flag is already set, the function returns immediately.  The return
 * value indicates the state of flag after waiting.
 *
 * If abs_time is NULL, e_flag_timed_wait() acts like e_flag_wait().
 *
 * To easily calculate abs_time, a combination of g_get_current_time() and
 * g_time_val_add() can be used.
 *
 * Returns TRUE if flag is now set
 */
extern gboolean	e_flag_timed_wait(EFlag *flag,
				  GTimeVal *abs_time);

/*
 * Destroys flag.
 */
extern void e_flag_free(EFlag *flag);

G_END_DECLS

#endif /* E_FLAG_H */
