/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Implementation file for a Brutus proxy daemon.
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>

#include "brutusd.h"
#include "brutus-proxy_impl.c"

#ifndef ORBIT2_EX
#define ORBIT2_EX(_ev_) ((NULL != _ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION))
#endif

static const char *prog_name = NULL;
static CORBA_ORB global_orb = CORBA_OBJECT_NIL;
static int global_lock_fd = -1;
static int ref_file_fd = -1;

static inline size_t
corba_strlen(const CORBA_char *s)
{
	size_t l = 0;

	for (; (CORBA_char)'\0' != *s; s++, l++)
		;

	return l;
}

// shutdown global ORB
static void
orb_shutdown(int sig)
{
	if (CORBA_OBJECT_NIL != global_orb) {
		CORBA_Environment ev[1];

		CORBA_exception_init(ev);
		CORBA_ORB_shutdown(global_orb, TRUE, ev);
		CORBA_exception_free(ev);
	}
	global_orb = CORBA_OBJECT_NIL;
}

static CORBA_ORB
create_orb(CORBA_Environment *ev)
{
	CORBA_ORB orb = CORBA_OBJECT_NIL;
	int init_argc = 5;
	char **init_argv = NULL;

#ifdef HAVE_ORBIT2_TIMEOUT
	init_argc++;
#else
#ifdef HAVE_ORBIT2_TIMEOUT_NG
	init_argc++;
#endif
#endif
	init_argv = (char**)malloc(sizeof(char*) * (init_argc));
	if (!init_argv)
		return CORBA_OBJECT_NIL;

	init_argv[0] = "DUMMY_ARGV0";

	// IPv4 - needed to interoperate with TAO
	init_argv[1] = "--ORBIIOPIPv4=1";

	// IPv6 - could be needed to interoperate with TAO
	init_argv[2] = "--ORBIIOPIPv6=1";

	// Explicitly force ORBit2 to be non-local
	init_argv[3] = "--ORBLocalOnly=0";

	// Force ORBit2 to use the IP address in the all generated IORs
	init_argv[4] = "--ORBNetID=ipaddr";

	/* 15 seconds timeout */
#ifdef HAVE_ORBIT2_TIMEOUT /* in microseconds */
	init_argv[5] = "--GIOPTimeoutLimit=15000000";
#else
#ifdef HAVE_ORBIT2_TIMEOUT_NG /* in milliseconds */
	init_argv[5] = "--GIOPTimeoutMSEC=15000";
#endif
#endif
	// initialize the ORB
	orb = CORBA_ORB_init(&init_argc, init_argv, "proxy_orb_orbit-io-thread", ev);
	if (ORBIT2_EX(ev))
		orb = CORBA_OBJECT_NIL;

	free(init_argv);

	return orb;
}

static int
object_ref_to_file(CORBA_ORB orb,
		   CORBA_Object object,
		   gchar *filename,
		   CORBA_Environment *ev)
{
	gboolean retv = FALSE;
	CORBA_char *objref = NULL;
	int file_fd = -1;
	FILE *file = NULL;
	size_t c = 0;

	file_fd = g_open(filename, O_CREAT | O_WRONLY | O_SYNC | O_TRUNC, S_IRUSR | S_IWUSR);
	if (-1 == file_fd)
		return -1;

	file = fdopen(file_fd, "w");
	if (!file)
		goto out;

	objref = CORBA_ORB_object_to_string(orb, object, ev);
	if (ORBIT2_EX(ev) || !objref)
		goto out;

	// must work even if (sizeof(char) != sizeof(CORBA_char))
	c = fwrite((const void*)objref, sizeof(CORBA_char), corba_strlen(objref)+1, file);
	if (c == (1 + corba_strlen(objref))) {
		retv = TRUE;
		fsync(file_fd);
	}

out:
	if (objref)
		CORBA_free(objref);

	if (file) {
		if (fclose(file))
			retv = FALSE;
	} else {
		if (close(file_fd))
			retv = FALSE;
	}
	if (!retv)
		return -1;

	return g_open(filename, O_CREAT | O_RDONLY | O_SYNC, S_IRUSR | S_IWUSR);
}

static CORBA_Object
brutus_object_ref_from_file(CORBA_ORB orb,
			    const gchar *filename,
			    CORBA_Environment *ev)
{
	CORBA_Object obj = CORBA_OBJECT_NIL;
	CORBA_char *objref = NULL;
	FILE *file = NULL;
	int fd = -1;
	struct stat st;
	size_t c = 0;

	file = g_fopen(filename, "r");
	if (!file)
		return CORBA_OBJECT_NIL;
	fd = fileno(file);
	if (-1 == fd)
		goto out;

	if (g_stat(filename, &st))
		goto out;

	objref = g_malloc0(st.st_size);
	if (!objref)
		goto out;

	// must work even if sizeof(char) != sizeof(CORBA_char)
	c = fread((void*)objref, sizeof(CORBA_char), (size_t)(st.st_size/sizeof(CORBA_char)), file);
	if (c != (size_t)(st.st_size/sizeof(CORBA_char)))
		goto out;

	obj = (CORBA_Object)CORBA_ORB_string_to_object(orb, objref, ev);

out:
	if (objref)
		g_free(objref);
	fclose(file);

	return obj;
}

static Lorica_ReferenceMapper
get_lorica(CORBA_Environment *ev)
{
	CORBA_boolean cb = CORBA_TRUE;
	Lorica_ReferenceMapper mapper = CORBA_OBJECT_NIL;

	// get lorica reference mapper if wanted and available locally
	if (!g_file_test(LORICA_DEFAULT_IOR_FILE, G_FILE_TEST_IS_REGULAR))
		goto no_lorica;

	mapper = (Lorica_ReferenceMapper)brutus_object_ref_from_file(global_orb, 
								     LORICA_DEFAULT_IOR_FILE, 
								     ev);
	if (ORBIT2_EX(ev))
		goto no_lorica;
	cb = CORBA_Object_is_nil((CORBA_Object)mapper, ev);
	if (ORBIT2_EX(ev))
		goto no_lorica;
	if (cb)
		goto no_lorica;
	goto lorica_ok;

no_lorica:
	mapper = CORBA_OBJECT_NIL;

lorica_ok:

	return mapper;
}

typedef enum {
	EXIT_DAEMON = 0, /* we are the daemon                  */
	EXIT_OK = 1,	 /* caller must exit with EXIT_SUCCESS */
	EXIT_ERROR = 2,	 /* caller must exit with EXIT_FAILURE */
} daemon_exit_t;

static daemon_exit_t
become_daemon(void)
{
	struct sigaction sig_act;
	struct rlimit rl;
	unsigned int n;
	pid_t pid = -1;
	int fd0;
	int fd1;
	int fd2;

	/*
	 * A process that has terminated but has not yet been waited for is a zombie.
	 * On the other hand, if the parent dies first, init (process 1) inherits the
	 * child and becomes its parent.
	 *
	 * (*) Citation from <http://www.win.tue.nl/~aeb/linux/lk/lk-10.html>
	 */

	/* fork off the parent process to create the session daemon */
	pid = fork();
	switch (pid) {
	case -1 :
		return EXIT_ERROR;
	case 0 :	
		/* We are the intermediate child.
		 * Now fork once more and try to ensure that this intermediate child
		 * exits before the final child so that the final child is adopted
		 * by init and does not become a zombie. This is racy as the final
		 * child concievably could terminate (and become a zombie) before
		 * this child exits. Knock on wood...
		 */

		if ((setsid()) < 0)
			return EXIT_ERROR;

		sig_act.sa_handler = SIG_IGN;
		sigemptyset(&sig_act.sa_mask);
		sig_act.sa_flags = 0;
		if (sigaction(SIGHUP, &sig_act, NULL))
			return EXIT_ERROR;

		pid = fork();
		switch (pid) {
		case -1 :
			return EXIT_ERROR;
		case 0 :
			break;
		default :
			return EXIT_OK;
		}

		/*
		 * (0 == pid) we are the final child
		 */

		/* sleep for 1 second to give the parent plenty of time to exit */
		g_usleep(1000000);

		break;
	default :
		/* wait for intermediate child */
		waitpid(pid, NULL, 0);

		return EXIT_OK;
	}

	/*
	 * We are now effectively the daemon and must continue
	 * to prep the daemon process for operation
	 */

	// change the working directory
	if ((chdir("/")) < 0)
		return EXIT_ERROR;

	// close any and all open file descriptors
	if (getrlimit(RLIMIT_NOFILE, &rl))
		return EXIT_ERROR;
	if (RLIM_INFINITY == rl.rlim_max)
		rl.rlim_max = 1024;
	for (n = 0; n < rl.rlim_max; n++) {
		if (close(n) && (EBADF != errno))
			return EXIT_ERROR;
	}

	// attach file descriptors 0, 1 and 2 to /dev/null
	fd0 = open("/dev/null", O_RDWR);
	fd1 = dup2(fd0, 1);
	fd2 = dup2(fd0, 2);
	if (0 != fd0)
		return EXIT_ERROR;

	return EXIT_DAEMON;
}

static int 
lock_fd(const int fd)
{
	struct flock lock = {
		.l_type = F_WRLCK,
		.l_start = 0,
		.l_whence = SEEK_SET,
		.l_len = 0,
	};

	if (-1 == fd)
		return -1;

	return fcntl(fd, F_SETLK, &lock);
}

static int
unlock_fd(const int fd)
{
	struct flock lock = {
		.l_type = F_UNLCK,
		.l_start = 0,
		.l_whence = SEEK_SET,
		.l_len = 0,
	};

	if (-1 == fd)
		return -1;

	return fcntl(fd, F_SETLK, &lock);
}

/*
 * Will attempt to lock the PID file and write the
 * pid into it. This function will return FALSE for
 * all errors. Callee is responsible for closing *fd 
 * if (*fd != -1).
 */
static gboolean
get_process_lock(int *fd)
{
	char buf[32] = { '\0' };
	int p = 0;
	ssize_t w = 0;
	char *file_path = NULL;

	*fd = -1;

	file_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_PID_FILE_NAME, NULL);
	if (!file_path)
		return FALSE;

	*fd = g_open(file_path, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	g_free(file_path);
	if (-1 == *fd)
		return FALSE;

	// lock it
	if (lock_fd(*fd))
		return FALSE;

	// we have the lock
	if (ftruncate(*fd, 0))
		return FALSE;

	/* write pid */
	p = snprintf(buf, sizeof(buf), "%d", getpid());
	w = write(*fd, buf, strlen(buf) + sizeof(char));
	p += sizeof(char);
	if (p != (int)w)
		return FALSE;

	if (fdatasync(*fd))
		return FALSE;

	return TRUE;
}

/*
 * This handler will only be called after the signal handler
 * is installed. The signal handler is not installed until
 * the lock on th epid file is in place.
 */
static void
restart(int sig)
{
	unlock_fd(ref_file_fd); // unsafe in signal handler
	close(ref_file_fd);
	ref_file_fd = -1;

	unlock_fd(global_lock_fd); // unsafe in signal handler
	close(global_lock_fd);
	global_lock_fd = -1;

	if (-1 == system(prog_name)) // unsafe in signal handler
		; // to avoid comiler warning
	orb_shutdown(sig); // unsafe in signal handler
}

int
main(int argc, char *argv[])
{
	PortableServer_POA root_poa = CORBA_OBJECT_NIL;
	BRUTUS_BrutusProxy proxy = CORBA_OBJECT_NIL;
	Lorica_ReferenceMapper mapper = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];
	gboolean daemon = TRUE;
	int retv = EXIT_FAILURE;
	gchar *ref_file_path = NULL;
	sigset_t mask;
	struct sigaction sig_act;
	daemon_exit_t dstat = EXIT_DAEMON;

	/* reset the file mode mask */
	umask(0);

	if (1 < argc)
		daemon = FALSE;

	/* restore signals */
	sigfillset(&mask);
	pthread_sigmask(SIG_UNBLOCK, &mask, NULL);

	/* for restart */
	prog_name = argv[0];

	/* create application directory */
	{
		gchar *dir_path = NULL;
		int retv = 1;

		dir_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_DIR_NAME, NULL);
		if (!dir_path)
			exit(EXIT_FAILURE);

		retv = g_mkdir_with_parents(dir_path, 0700);
		g_free(dir_path);
		if (retv)
			exit(EXIT_FAILURE);
	}

	if (daemon)
		dstat = become_daemon();
	switch (dstat) {
	case EXIT_DAEMON :
		break;
	case EXIT_OK :
		exit(EXIT_SUCCESS);
	case EXIT_ERROR :
	default :
		exit(EXIT_FAILURE);
	}

	/*
	 * We are now the would-be log daemon
	 */

	// take the lock and write the pid
	if (!get_process_lock(&global_lock_fd)) {
		if (-1 != global_lock_fd) {
			close(global_lock_fd);
			global_lock_fd = -1;
		}
		exit(EXIT_SUCCESS);
	}

	/*
	 * We are now ensured that we are the only log daemon in the air
	 * unless HOME is on NFS...
	 *
	 * All exits must be taken by "goto out;".
	 */

	CORBA_exception_init(ev);

	/*
	 * Initialize CORBA
	 */

	global_orb = create_orb(ev);
	if (ORBIT2_EX(ev))
		goto out;
	{
		PortableServer_POAManager poa_manager = CORBA_OBJECT_NIL;

		root_poa = (PortableServer_POA) CORBA_ORB_resolve_initial_references(global_orb, "RootPOA", ev);
		if (ORBIT2_EX(ev))
			goto out;

		poa_manager = PortableServer_POA__get_the_POAManager(root_poa, ev);
		if (ORBIT2_EX(ev))
			goto out;

		PortableServer_POAManager_activate(poa_manager, ev);
		if (ORBIT2_EX(ev))
			goto out;

		CORBA_Object_release((CORBA_Object)poa_manager, ev);
		if (ORBIT2_EX(ev))
			goto out;
	}

	/* get Lorica if available */
	mapper = get_lorica(ev);

	/* create proxy servant */
	proxy = impl_BRUTUS_BrutusProxy__create(root_poa, mapper, ev);
	if (ORBIT2_EX(ev))
		goto out;
	if (CORBA_OBJECT_NIL == proxy)
		goto out;

	/* write ior to file */
	ref_file_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_REF_FILE_NAME, NULL);
	if (!ref_file_path)
		goto out;
	ref_file_fd = object_ref_to_file(global_orb, proxy, ref_file_path, ev);
	if (-1 == ref_file_fd)
		goto out;

	/*
	 * handle signals
	 */

	sig_act.sa_handler = orb_shutdown;
	sigfillset(&sig_act.sa_mask); // ignore as many signals as possible when in the handler!
	sig_act.sa_flags = SA_RESETHAND;
	sigaction(SIGUSR1, &sig_act, NULL);

	/* restart the daemon if SIGUSR2 is caught */
	if (daemon)
		sig_act.sa_handler = restart;
	else
		sig_act.sa_handler = orb_shutdown;
	sigfillset(&sig_act.sa_mask);
	sig_act.sa_flags = SA_RESETHAND;
	sigaction(SIGUSR2, &sig_act, NULL);

	sig_act.sa_handler = orb_shutdown;
	sigfillset(&sig_act.sa_mask);
	sig_act.sa_flags = SA_RESETHAND;
	sigaction(SIGTERM, &sig_act, NULL);

	sig_act.sa_handler = orb_shutdown;
	sigfillset(&sig_act.sa_mask);
	sig_act.sa_flags = SA_RESETHAND;
	sigaction(SIGINT, &sig_act, NULL);

	/* loop until killed */
	lock_fd(ref_file_fd);
	CORBA_ORB_run(global_orb, ev);
	unlock_fd(ref_file_fd);

	/* cleanup */
	{
		PortableServer_ObjectId *objid = NULL;

		CORBA_Object_release((CORBA_Object)mapper, ev);
		if (ORBIT2_EX(ev))
			goto out;

		objid = PortableServer_POA_reference_to_id(root_poa, (CORBA_Object)proxy, ev);
		if (ORBIT2_EX(ev)) {
			if (objid)
				CORBA_free(objid);
			goto out;
		}

		PortableServer_POA_deactivate_object(root_poa, objid, ev);
		CORBA_free(objid);
		if (ORBIT2_EX(ev))
			goto out;

		CORBA_Object_release((CORBA_Object)proxy, ev);
		if (ORBIT2_EX(ev))
			goto out;

		PortableServer_POA_destroy(root_poa, TRUE, FALSE, ev);
		if (ORBIT2_EX(ev))
			goto out;

		CORBA_Object_release((CORBA_Object)root_poa, ev);
		if (ORBIT2_EX(ev))
			goto out;

		/* ORB: tear down the ORB */
		if (CORBA_OBJECT_NIL != global_orb) {
			CORBA_ORB_destroy(global_orb, ev);
			if (ORBIT2_EX(ev))
				goto out;

			CORBA_Object_release((CORBA_Object)global_orb, ev);
			if (ORBIT2_EX(ev))
				goto out;
			global_orb = CORBA_OBJECT_NIL;
		}
	}
	retv = EXIT_SUCCESS;

out:
	/*
	 * We only ever end up here if we succefully took
	 * the lock and became the daemon
	 */

	if (CORBA_OBJECT_NIL != mapper)
		CORBA_Object_release((CORBA_Object)mapper, ev); 

	CORBA_exception_free(ev);

	/*
	 * There is a small race where the daemon could catch a signal just after
	 * installing signal handlers but before placing the lock. Releasing the
	 * lock unnecessarily should not harm us though.
	 */
	if (-1 != ref_file_fd)
		close(ref_file_fd);

	if (ref_file_path) {
		if (-1 == truncate(ref_file_path, 0))
			; // to avoid compiler warning
		g_free(ref_file_path);
	}

	if (-1 != global_lock_fd) {
		unlock_fd(global_lock_fd);
		close(global_lock_fd);
	}

	exit(retv);
}
